from collections.abc import Callable

from jaxtyping import PyTree

from .._eval_shape import filter_eval_shape
from .._module import Module
from .._tree import tree_at, tree_equal


class SharedNode:
    """Placeholder value for nodes that have been removed by `eqx.nn.Shared`."""

    def __repr__(self):
        return "SharedNode"


class Shared(Module, strict=True):
    """Used to tie together multiple nodes across a PyTree.

    Note that Equinox modules are Py**Trees** -- so the same layer, appearing in two
    difference parts of the tree, will be treated as two copies of this layer. For
    example,
    ```python
    class SubModel(eqx.Module):
        linear: eqx.nn.Linear

    class Model(eqx.Module):
        linear: eqx.nn.Linear
        submodel: SubModel

        def __init__(self):
            linear = eqx.nn.Linear(...)
            self.linear = linear
            self.submodel = SubModel(linear)
    ```
    is used to declare `model.linear` and `model.submodel.linear` as two separate
    layers. They will start with the same initial parameter values, and then update
    independently during training.

    For when we really do want to share layers or weights across different parts of a
    model, then `eqx.nn.Shared` exists as a way to easily express this in the PyTree
    paradigm.

    !!! Example

        It is common in many language models to have an initial embedding matrix at the
        start, and then to reuse this as the weight of the final linear transformation.

        ```python
        import equinox as eqx
        import jax.numpy as jnp
        from jaxtyping import Array, Int

        class LanguageModel(eqx.Module):
            shared: eqx.nn.Shared

            def __init__(self):
                embedding = eqx.nn.Embedding(...)
                linear = eqx.nn.Linear(...)
                # These two weights will now be tied together.
                where = lambda embed_and_lin: embed_and_lin[1].weight
                get = lambda embed_and_lin: embed_and_lin[0].weight
                self.shared = eqx.nn.Shared((embedding, linear), where, get)

            def __call__(self, tokens: Int[Array, "sequence"]):
                # Expand back out so we can evaluate these layers.
                embedding, linear = self.shared()
                assert embedding.weight is linear.weight  # same parameter!
                # Now go ahead and evaluate your language model.
                values = jax.vmap(embedding)(tokens)
                ...  # other layers, probably
                return jax.vmap(linear)(values)
        ```

        _(Side note: you will sometimes see some authors referring to transposing
        the embedding matrix prior to the final linear layer. This is because some
        other libraries store the weight matrices of linear layers the other way
        around. If that had been necessary here then we could have done it with
        `get = lambda embed_and_lin: jnp.transpose(embed_and_lin[0].weight)`.)_

    """

    pytree: PyTree
    where: Callable
    get: Callable

    def __init__(self, pytree: PyTree, where: Callable, get: Callable):
        """**Arguments:**

        - `pytree`: The PyTree to share some nodes across.
        - `where`: a function specifying either a single node, or a sequence of nodes,
            as with `eqx.tree_at(where, pytree, ...)`.
        - `get`: a function, which when evaluated on `pytree`, returns either a single
            value (if `where` does), or a sequence of values (if `where` does, and in
            this case this must be a sequence of the same length as `where`).

        The node(s) of `get(pytree)` and the corresponding value(s) of `where(pytree)`
        will be tied together.

        !!! info

            To explain how this works. The implementation is just:
            ```python
            class Shared(eqx.Module):
                pytree: PyTree
                where: Callable
                get: Callable

                def __init__(self, pytree, where, get):
                    # `0` is just some dummy value
                    self.pytree = eqx.tree_at(where, pytree, replace_fn=lambda _: 0)
                    self.where = where
                    self.get = get

                def __call__(self):
                    return eqx.tree_at(self.where, self.pytree, self.get(self.pytree))
            ```
            so that at `__init__` time, the duplicate nodes specified in `where` are
            removed from the PyTree. We no longer have a separate copy updating during
            training.

            And then at `__call__` time, references to the values returned by
            `get(pytree)` are put in their place. We end up with a pytree of the same
            structure as what we started with, which we can now use (evaluate as a
            layer etc.) as normal.

        !!! tip

            If you need to apply any transform (e.g. transposing a matrix), then this
            can be done as part of `get`. For example,
            `get = lambda pair: jnp.transpose(pair[1].weight)`.
        """

        source_struct = filter_eval_shape(get, pytree)
        dest_struct = filter_eval_shape(where, pytree)
        if tree_equal(source_struct, dest_struct) is not True:
            raise ValueError(
                "Every node being shared together must have the same pytree "
                "structure, shape+dtype of arrays, etc., as each other. Got:\n"
                f"{source_struct}\n"
                "and\n"
                f"{dest_struct}"
            )
        self.pytree = tree_at(where, pytree, replace_fn=lambda _: SharedNode())
        self.where = where
        self.get = get

    def __call__(self):
        """**Arguments:**

        None.

        **Returns:**

        A PyTree of the same structure as the original `pytree`, with `get(pytree)` in
        the place of the nodes at `where(pytree)`.
        """
        return tree_at(self.where, self.pytree, self.get(self.pytree))



from collections.abc import Hashable, Sequence
from typing import Optional, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, PRNGKeyArray

from .._misc import default_floating_dtype
from .._module import field
from ._sequential import StatefulLayer
from ._stateful import State, StateIndex


class BatchNorm(StatefulLayer, strict=True):
    r"""Computes a mean and standard deviation over the batch and spatial
    dimensions of an array, and uses these to normalise the whole array. Optionally
    applies a channelwise affine transformation afterwards.

    Given an input array $x = [x_1, ... x_C]$ with $C$ channels, this layer computes

    $$\frac{x_i - \mathbb{E}[x_i]}{\sqrt{\text{Var}[x_i] + \varepsilon}} * \gamma_i + \beta_i$$

    for all $i$. Here $*$ denotes elementwise multiplication and $\gamma$, $\beta$ have
    shape $(C,)$ if `channelwise_affine=True` and $\gamma = 1$, $\beta = 0$ if
    `channelwise_affine=False`. Expectations are computed over all spatial dimensions
    *and* over the batch dimension, and updated batch-by-batch according to `momentum`.

    !!! example

        See [this example](../../examples/stateful.ipynb) for example usage.

    !!! warning

        This layer must be used inside of a `vmap` or `pmap` with a matching
        `axis_name`. (Not doing so will raise a `NameError`.)

    Note that this layer behaves differently during training and inference. During
    training then statistics are computed using the input data, and the running
    statistics updated. During inference then just the running statistics are used.
    Whether the model is in training or inference mode should be toggled using
    [`equinox.nn.inference_mode`][].
    """  # noqa: E501

    weight: Optional[Float[Array, "input_size"]]
    bias: Optional[Float[Array, "input_size"]]
    first_time_index: StateIndex[Bool[Array, ""]]
    state_index: StateIndex[
        tuple[Float[Array, "input_size"], Float[Array, "input_size"]]
    ]
    axis_name: Union[Hashable, Sequence[Hashable]]
    inference: bool
    input_size: int = field(static=True)
    eps: float = field(static=True)
    channelwise_affine: bool = field(static=True)
    momentum: float = field(static=True)

    def __init__(
        self,
        input_size: int,
        axis_name: Union[Hashable, Sequence[Hashable]],
        eps: float = 1e-5,
        channelwise_affine: bool = True,
        momentum: float = 0.99,
        inference: bool = False,
        dtype=None,
    ):
        """**Arguments:**

        - `input_size`: The number of channels in the input array.
        - `axis_name`: The name of the batch axis to compute statistics over, as passed
            to `axis_name` in `jax.vmap` or `jax.pmap`. Can also be a sequence (e.g. a
            tuple or a list) of names, to compute statistics over multiple named axes.
        - `eps`: Value added to the denominator for numerical stability.
        - `channelwise_affine`: Whether the module has learnable channel-wise affine
            parameters.
        - `momentum`: The rate at which to update the running statistics. Should be a
            value between 0 and 1 exclusive.
        - `inference`: If `False` then the batch means and variances will be calculated
            and used to update the running statistics. If `True` then the running
            statistics are directly used for normalisation. This may be toggled with
            [`equinox.nn.inference_mode`][] or overridden during
            [`equinox.nn.BatchNorm.__call__`][].
        - `dtype`: The dtype to use for the running statistics and the weight and bias
            if `channelwise_affine` is `True`. Defaults to either
            `jax.numpy.float32` or `jax.numpy.float64` depending on whether JAX is in
            64-bit mode.
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        if channelwise_affine:
            self.weight = jnp.ones((input_size,), dtype=dtype)
            self.bias = jnp.zeros((input_size,), dtype=dtype)
        else:
            self.weight = None
            self.bias = None
        self.first_time_index = StateIndex(jnp.array(True))
        init_buffers = (
            jnp.empty((input_size,), dtype=dtype),
            jnp.empty((input_size,), dtype=dtype),
        )
        self.state_index = StateIndex(init_buffers)
        self.inference = inference
        self.axis_name = axis_name
        self.input_size = input_size
        self.eps = eps
        self.channelwise_affine = channelwise_affine
        self.momentum = momentum

    @jax.named_scope("eqx.nn.BatchNorm")
    def __call__(
        self,
        x: Array,
        state: State,
        *,
        key: Optional[PRNGKeyArray] = None,
        inference: Optional[bool] = None,
    ) -> tuple[Array, State]:
        """**Arguments:**

        - `x`: A JAX array of shape `(input_size, dim_1, ..., dim_N)`.
        - `state`: An [`equinox.nn.State`][] object (which is used to store the
            running statistics).
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)
        - `inference`: As per [`equinox.nn.BatchNorm.__init__`][]. If
            `True` or `False` then it will take priority over `self.inference`. If
            `None` then the value from `self.inference` will be used.

        **Returns:**

        A 2-tuple of:

        - A JAX array of shape `(input_size, dim_1, ..., dim_N)`.
        - An updated state object (storing the updated running statistics).

        **Raises:**

        A `NameError` if no `vmap`s are placed around this operation, or if this vmap
        does not have a matching `axis_name`.
        """

        if inference is None:
            inference = self.inference
        if inference:
            running_mean, running_var = state.get(self.state_index)
        else:

            def _stats(y):
                mean = jnp.mean(y)
                mean = lax.pmean(mean, self.axis_name)
                var = jnp.mean((y - mean) * jnp.conj(y - mean))
                var = lax.pmean(var, self.axis_name)
                var = jnp.maximum(0.0, var)
                return mean, var

            first_time = state.get(self.first_time_index)
            state = state.set(self.first_time_index, jnp.array(False))

            batch_mean, batch_var = jax.vmap(_stats)(x)
            running_mean, running_var = state.get(self.state_index)
            momentum = self.momentum
            running_mean = (1 - momentum) * batch_mean + momentum * running_mean
            running_var = (1 - momentum) * batch_var + momentum * running_var
            running_mean = lax.select(first_time, batch_mean, running_mean)
            running_var = lax.select(first_time, batch_var, running_var)
            state = state.set(self.state_index, (running_mean, running_var))

        def _norm(y, m, v, w, b):
            out = (y - m) / jnp.sqrt(v + self.eps)
            if self.channelwise_affine:
                out = out * w + b
            return out

        out = jax.vmap(_norm)(x, running_mean, running_var, self.weight, self.bias)
        return out, state



from typing import Any, Optional

import jax
import jax.numpy as jnp
import jax.random as jrandom
from jax._src.dtypes import TypePromotionError
from jaxtyping import Array, ArrayLike, Float, Int, PRNGKeyArray

from .._caches import cache_clears
from .._filters import is_array_like
from .._misc import default_floating_dtype
from .._module import field, Module


internal_rope_embedding_cache: dict[tuple[int, Any], tuple[Array, Array]] = {}
cache_clears.append(internal_rope_embedding_cache.clear)


class Embedding(Module, strict=True):
    """A simple lookup table that stores embeddings of a fixed size."""

    num_embeddings: int = field(static=True)
    embedding_size: int = field(static=True)
    weight: Array

    def __init__(
        self,
        num_embeddings: Optional[int] = None,  # pyright: ignore
        embedding_size: Optional[int] = None,  # pyright: ignore
        weight: Optional[Float[Array, "num_embeddings embedding_size"]] = None,
        dtype=None,
        *,
        key: Optional[PRNGKeyArray] = None,
    ):
        """**Arguments:**

        `Embedding` should be initialised with either:

        - `num_embeddings`: Size of embedding dictionary. Must be non-negative.
        - `embedding_size`: Size of each embedding vector. Must be non-negative.
        - `dtype`: The dtype to use for the embedding weights. Defaults to either
            `jax.numpy.float32` or `jax.numpy.float64` depending on whether JAX is in
            64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for initialisation
            of the embedding lookup table. (Keyword only argument.)

        Or:

        - `weight`: The embedding lookup table, of shape
            `(num_embeddings, embedding_size)`.
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        if weight is None:
            if num_embeddings is None or embedding_size is None or key is None:
                raise ValueError(
                    "Must provide `eqx.nn.Embedding(num_embeddings=..., "
                    "embedding_size=..., key=...)` if not providing the weight "
                    "directly."
                )
            if num_embeddings < 0:
                raise ValueError("num_embeddings must not be negative.")
            if embedding_size < 0:
                raise ValueError("embedding_size must not be negative.")
            self.weight = jrandom.normal(
                key, (num_embeddings, embedding_size), dtype=dtype
            )
        else:
            if weight.ndim != 2:
                raise ValueError(
                    "weight must have shape (num_embeddings, embedding_size)."
                )
            if num_embeddings is None:
                num_embeddings: int = weight.shape[0]
            if embedding_size is None:
                embedding_size: int = weight.shape[1]
            if weight.shape != (num_embeddings, embedding_size):
                raise ValueError(
                    "weight must have shape (num_embeddings, embedding_size)."
                )
            self.weight = weight
        self.num_embeddings = num_embeddings
        self.embedding_size = embedding_size

    @jax.named_scope("eqx.nn.Embedding")
    def __call__(
        self, x: Int[ArrayLike, ""], *, key: Optional[PRNGKeyArray] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The table index. Should be a scalar integer array.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(embedding_size,)`, from the x-th index of the embedding
        table.
        """
        if is_array_like(x) and jnp.shape(x) == ():
            return self.weight[x]
        else:
            raise ValueError(
                "`eqx.nn.Embedding()(x)` should be called with a scalar index `x`. "
                "Use `jax.vmap` if you would like to index with multiple values."
            )


class RotaryPositionalEmbedding(Module, strict=True):
    """A rotary positional encoding module, as described in the paper
    "RoFormer: Enhanced Transformer with Rotary Position Embedding". While this module
    can be used in any context, it is particularly useful for providing positional
    information to transformer models.

    !!! Example

        The following example demonstrates how to use `RotaryPositionalEmbedding` in
        a simple transformer model.

        ```python
        class TransformerBlock(eqx.Module):
            rope_embeddings: RotaryPositionalEmbedding

            def __init__(...):
                self.rope_embeddings = RotaryPositionalEmbedding(...)

            def __call__(...):
                def process_heads(
                    query_heads: Float[Array, "seq_length num_heads qk_size"],
                    key_heads: Float[Array, "seq_length num_heads qk_size"],
                    value_heads: Float[Array, "seq_length num_heads vo_size"]
                ) -> tuple[
                    Float[Array, "seq_length num_heads qk_size"],
                    Float[Array, "seq_length num_heads qk_size"],
                    Float[Array, "seq_length num_heads vo_size"]
                ]:
                    query_heads = jax.vmap(self.rope_embeddings,
                                           in_axes=1,
                                           out_axes=1)(query_heads)
                    key_heads = jax.vmap(self.rope_embeddings,
                                         in_axes=1,
                                         out_axes=1)(key_heads)

                    return query_heads, key_heads, value_heads

                x = self.mha_attention(... process_heads=process_heads)
                ...
        ```

    ??? cite

        [RoFormer: Enhanced Transformer with Rotary Position Embedding](https://arxiv.org/pdf/2104.09864)

        ```bibtex
        @misc{su2023roformer,
          title={RoFormer: Enhanced Transformer with Rotary Position Embedding},
          author={Jianlin Su and Yu Lu and Shengfeng Pan and Ahmed Murtadha and
              Bo Wen and Yunfeng Liu},
          year={2023},
          eprint={arXiv:2104.09864},
        }
        ```
    """

    embedding_size: int = field(static=True)
    theta: float = field(static=True, default=10_000.0)
    dtype: Any = field(static=True, default_factory=default_floating_dtype)

    def __check_init__(self):
        if self.embedding_size < 0:
            raise ValueError("`embedding_size` must not be negative.")
        if (self.embedding_size % 2) != 0:
            raise ValueError("`embedding_size` must be even.")

    @staticmethod
    def rotate_half(x: Float[Array, "seq_length embedding_size"]):
        d_2 = x.shape[-1] // 2
        return jnp.concatenate([-x[..., d_2:], x[..., :d_2]], axis=-1)

    @staticmethod
    def precompute_freqs_cis(
        embedding_size: int, end: int, theta: float, dtype: Any
    ) -> tuple[Float[Array, "end half_emb_size"], Float[Array, "end half_emb_size"]]:
        freqs = 1.0 / (
            theta
            ** (jnp.arange(0.0, embedding_size, 2)[jnp.newaxis, :] / embedding_size)
        )

        t = jnp.arange(float(end))
        freqs_outer = jnp.outer(t, freqs)

        # we assign the type at the very end to minimize the loss of precision
        return jnp.cos(freqs_outer).astype(dtype), jnp.sin(freqs_outer).astype(dtype)

    @jax.named_scope("eqx.nn.RotaryPositionalEmbedding")
    def __call__(
        self,
        x: Float[Array, "seq_length embedding_size"],
        *,
        key: Optional[PRNGKeyArray] = None,
    ) -> Float[Array, "seq_length embedding_size"]:
        """**Arguments:**

        - `x`: A JAX array of shape `(seq_length, embedding_size)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(seq_length, embedding_size)`, with the rotary positional
        encoding applied to the input.
        """

        seq_len, embedding_size = x.shape
        if embedding_size != self.embedding_size:
            raise ValueError(
                f"x.shape[-1] must match self.embedding_size, "
                f"but {x.shape[-1]} != {self.embedding_size}"
            )

        with jax.ensure_compile_time_eval():
            cache_key = (embedding_size, self.dtype)
            if cache_key not in internal_rope_embedding_cache:
                internal_rope_embedding_cache[cache_key] = self.precompute_freqs_cis(
                    embedding_size, seq_len, self.theta, self.dtype
                )

            freqs_cos, freqs_sin = internal_rope_embedding_cache[cache_key]
            freqs_seq_len, _ = freqs_cos.shape
            if seq_len > freqs_seq_len:
                internal_rope_embedding_cache[cache_key] = self.precompute_freqs_cis(
                    embedding_size, seq_len, self.theta, self.dtype
                )
                freqs_cos, freqs_sin = internal_rope_embedding_cache[cache_key]

            freqs_cos = freqs_cos[:seq_len]
            freqs_sin = freqs_sin[:seq_len]

        freqs_cos = jnp.tile(freqs_cos, (1, 2))
        freqs_sin = jnp.tile(freqs_sin, (1, 2))

        rotate_x = self.rotate_half(x)
        try:
            x_rope = (x * freqs_cos) + (rotate_x * freqs_sin)
        except TypePromotionError as e:
            inp_dtype = jnp.dtype(x.dtype)
            rope_dtype = jnp.dtype(self.dtype)
            raise TypePromotionError(
                f"The type of the passed value differs from the type "
                f"of the rotary embeddings ({inp_dtype} != {rope_dtype}), thus leading "
                "to a conflict when numpy_dtype_promotion is set to strict. To avoid "
                f"this error, either initialiaze RoPE module with {inp_dtype} "
                f"dtype, or explicitly cast the input argument to {rope_dtype}."
            ) from e
        return x_rope.astype(x.dtype)


RotaryPositionalEmbedding.__init__.__doc__ = """**Arguments:**

- `embedding_size`: Size of each embedding vector. Must be non-negative and even.
- `theta`: The base frequency for the sinusoidal functions used in positional encoding.
    Specifies how quickly the inner-product will decay with relative distance between
    tokens. Larger values of theta will result in slower oscillations. Default is
    10_000, as per the original paper.
- `dtype`: The dtype to use for the precomputed frequencies. Defaults to either
    `jax.numpy.float32` or `jax.numpy.float64` depending on whether JAX is in
    64-bit mode.
"""



from collections.abc import Callable
from typing import Any

import jax
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.tree_util as jtu

from .._filters import combine, is_array, partition


def announce_transform(
    x, name=None, intermediates=False, announce: Callable[[str], Any] = print
):
    """Identity function on an arbitrary PyTree. Announces each time a JAX transform is
    applied (grad, vmap, etc.).

    !!! warning

        This API is not stable. It should be used for one-off debugging purposes only.

    **Arguments:**

    - `x`: a variable to intercept.
    - `intermediates`: whether to include intermediate transforms, that haven't yet
        finished being transformed. E.g. if
        `intermediates=False`, then `jit(vmap(...))` will print out
        ```
        vmap:abstract
        vmap:mlir`
        ```
        whilst if `intermediates=True`, then `jit(vmap(...))` will print out
        ```
        vmap
        vmap:abstract
        vmap:mlir`
        ```
    - `announce`: the function to announce via. Defaults to just `print`.

    **Returns:**

    The `x` argument is returned unchanged.

    As a side-effect, the transforms applied to `x` will be printed out.
    """
    if name is None:
        name = ""
    else:
        name = f"({name}) "
    array, nonarray = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(array)
    flat = announce_jaxpr_p.bind(
        *flat, stack=(), name=name, intermediates=intermediates, announce=announce
    )
    array = jtu.tree_unflatten(treedef, flat)
    return combine(array, nonarray)


def _impl(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("impl",)
    announce(name + ":".join(stack))
    return x


def _abstract(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("abstract",)
    announce(name + ":".join(stack))
    return x


def _jvp(p, t, *, stack, name, intermediates, announce):
    if intermediates:
        _stack = stack + ("jvp",)
        announce(name + ":".join(_stack))
    p_stack = stack + ("jvp_p",)
    t_stack = stack + ("jvp_t",)
    p_out = announce_jaxpr_p.bind(
        *p, stack=p_stack, name=name, intermediates=intermediates, announce=announce
    )
    t_nonzero = [ti for ti in t if type(ti) is not ad.Zero]
    if len(t_nonzero) > 0:
        t_nonzero_out = announce_jaxpr_p.bind(
            *t_nonzero,
            stack=t_stack,
            name=name,
            intermediates=intermediates,
            announce=announce,
        )
    else:
        t_nonzero_out = []
    t_nonzero_out = iter(t_nonzero_out)
    t_out = [ti if type(ti) is ad.Zero else next(t_nonzero_out) for ti in t]
    assert next(t_nonzero_out, None) is None
    return p_out, t_out


def _transpose(ct, p, *, stack, name, intermediates, announce):
    stack = stack + ("transpose",)
    if intermediates:
        announce(name + ":".join(stack))
    return announce_jaxpr_p.bind(
        *ct, stack=stack, name=name, intermediates=intermediates, announce=announce
    )


def _batching(p, b, *, stack, name, intermediates, announce):
    stack = stack + ("vmap",)
    if intermediates:
        announce(name + ":".join(stack))
    out = announce_jaxpr_p.bind(
        *p, stack=stack, name=name, intermediates=intermediates, announce=announce
    )
    return out, b


def _mlir(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("mlir",)
    announce(name + ":".join(stack))
    return x


announce_jaxpr_p = jax.core.Primitive("announce_jaxpr")
announce_jaxpr_p.multiple_results = True
announce_jaxpr_p.def_impl(_impl)
announce_jaxpr_p.def_abstract_eval(_abstract)
ad.primitive_jvps[announce_jaxpr_p] = _jvp
ad.primitive_transposes[announce_jaxpr_p] = _transpose
batching.primitive_batchers[announce_jaxpr_p] = _batching
mlir.register_lowering(announce_jaxpr_p, mlir.lower_fun(_mlir, multiple_results=True))



from collections.abc import Callable, Sequence
from typing import Any, Optional, overload, Union

import jax
import jax.random as jr
from jaxtyping import Array, PRNGKeyArray

from .._better_abstract import AbstractClassVar
from .._custom_types import sentinel
from .._module import Module, StrictConfig
from ._stateful import State


class StatefulLayer(Module, strict=StrictConfig(allow_abstract_name=True)):
    """An abstract base class, used by [`equinox.nn.Sequential`][], to mark that a
    layer might be stateful. If `Sequential` sees that a layer inherits from
    `StatefulLayer`, then it will call `layer.is_stateful()` to check whether to
    call the layer as `new_x = layer(x)` or `(new_x, new_state) = layer(x, state)`.
    """

    def is_stateful(self) -> bool:
        """Indicates whether this layer should be considered stateful.

        The default implementation just returns True, but subclasses may override
        this to provide custom logic if the layer is only "maybe stateful". (E.g. if
        they optionally use stateful sublayers themselves.)

        **Arguments:**

        None

        **Returns:**

        A boolean. `True` indicates that the layer should be called as
        `(new_x, new_state) = layer(x, state)`. `False` indicates that the layer
        should be called as `new_x = layer(x)`.
        """
        return True

    __call__: AbstractClassVar[Callable]


class Sequential(StatefulLayer, strict=StrictConfig(allow_method_override=True)):
    """A sequence of [`equinox.Module`][]s applied in order.

    !!! note

        Activation functions can be added by wrapping them in [`equinox.nn.Lambda`][].
    """

    layers: tuple

    def __init__(self, layers: Sequence[Callable]):
        """**Arguments:**

        - `layers`: A sequence of [`equinox.Module`][]s.
        """

        self.layers = tuple(layers)

    def is_stateful(self) -> bool:
        return any(
            isinstance(x, StatefulLayer) and x.is_stateful() for x in self.layers
        )

    @overload
    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array: ...

    @overload
    def __call__(
        self, x: Array, state: State, *, key: Optional[PRNGKeyArray] = None
    ) -> tuple[Array, State]: ...

    @jax.named_scope("eqx.nn.Sequential")
    def __call__(
        self,
        x: Array,
        state: State = sentinel,
        *,
        key: Optional[PRNGKeyArray] = None,
    ) -> Union[Array, tuple[Array, State]]:
        """**Arguments:**

        - `x`: passed to the first member of the sequence.
        - `state`: If provided, then it is passed to, and updated from, any layer
            which subclasses [`equinox.nn.StatefulLayer`][].
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**
        The output of the last member of the sequence.

        If `state` is passed, then a 2-tuple of `(output, state)` is returned.
        If `state` is not passed, then just the output is returned.
        """

        if key is None:
            keys = [None] * len(self.layers)
        else:
            keys = jr.split(key, len(self.layers))
        for layer, key in zip(self.layers, keys):
            if isinstance(layer, StatefulLayer) and layer.is_stateful():
                x, state = layer(x, state=state, key=key)
            else:
                x = layer(x, key=key)
        if state is sentinel:
            return x
        else:
            return x, state

    def __getitem__(self, i: Union[int, slice]) -> Callable:
        if isinstance(i, int):
            return self.layers[i]
        elif isinstance(i, slice):
            return Sequential(self.layers[i])
        else:
            raise TypeError(f"Indexing with type {type(i)} is not supported")

    def __iter__(self):
        yield from self.layers

    def __len__(self):
        return len(self.layers)


class Lambda(Module, strict=True):
    """Wraps a callable (e.g. an activation function) for use with
    [`equinox.nn.Sequential`][].

    Precisely, this just adds an extra `key` argument (that is ignored). Given some
    function `fn`, then `Lambda` is essentially a convenience for `lambda x, key: f(x)`.

    !!! faq

        If you get a TypeError saying the function is not a valid JAX type, see the
            [FAQ](https://docs.kidger.site/equinox/faq/).

    !!! Example

        ```python
           model = eqx.nn.Sequential(
               [
                   eqx.nn.Linear(...),
                   eqx.nn.Lambda(jax.nn.relu),
                   ...
               ]
           )
        ```
    """

    fn: Callable[[Any], Any]

    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array:
        """**Arguments:**

        - `x`: The input JAX array.
        - `key`: Ignored.

        **Returns:**

        The output of the `fn(x)` operation.
        """
        return self.fn(x)


Lambda.__init__.__doc__ = """**Arguments:**

- `fn`: A callable to be wrapped in [`equinox.Module`][].
"""
